
class Stage(object):
    '''Contains a collection of stage related methods.'''

    async def create_stage_instance(self, channel_id, topic: str, privacy_level: int=None) -> dict:
        '''https://discord.com/developers/docs/resources/stage-instance#create-stage-instance'''
        payload: dict = {
            'channel_id': channel_id,
            'topic': topic,
            'privacy_level': privacy_level
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await self._request('POST', params=payload, uri=f'/stage-instances')

    async def get_stage_instance(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/stage-instance#get-stage-instance'''
        return await self._request('GET', uri=f'/stage-instances/{channel_id}')

    async def modify_stage_instance(self, topic: str=None, privacy_level=None) -> dict:
        '''https://discord.com/developers/docs/resources/stage-instance#modify-stage-instance'''
        payload: dict = { 'topic': topic, 'privacy_level': privacy_level }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await self._request('PATCH', params=payload, uri=f'/stage-instances/{channel_id}')

    async def delete_stage_instance(self) -> dict:
        '''https://discord.com/developers/docs/resources/stage-instance#delete-stage-instance'''
        return await self._request('DELETE', uri=f'/stage-instances/{channel_id}')


import numpy as np


class Intents(object):
    '''Object to store and serve intents for WSClient '''

    _intents = [
        'GUILDS', 'GUILD_MEMBERS', 'GUILD_BANS', 'GUILD_EMOJIS_AND_STICKERS', 'GUILD_INTEGRATIONS', 'GUILD_WEBHOOKS',
        'GUILD_INVITES', 'GUILD_VOICE_STATES', 'GUILD_PRESENCES', 'GUILD_MESSAGES', 'GUILD_MESSAGE_REACTIONS',
        'GUILD_MESSAGE_TYPING', 'DIRECT_MESSAGES', 'DIRECT_MESSAGE_REACTIONS', 'GUILD_SCHEDULED_EVENTS', 'GUILD_SCHEDULED_EVENTS'
    ]

    @classmethod
    def get_intents_list(cls) -> [str]:
        return cls._intents

    def get_intent(self, intent_name: str='') -> int:
        for i, role in enumerate(self._intents):
            if role == intent_name:
                return (1<<i)
        raise ValueError(f'Intent {intent_name} not found!')
    
    def _get_intents(self, intents_list: [str]) -> int:
        return np.sum([self.get_intent(intent) for intent in intents_list])

    
    # def _get_intents(self, intents: str) -> int:
    #     # https://discord.com/developers/docs/topics/gateway#gateway-intents
    #     if intents == 'GUILDS':
    #         '''
    #             - GUILD_CREATE
    #             - GUILD_UPDATE
    #             - GUILD_DELETE
    #             - GUILD_ROLE_CREATE
    #             - GUILD_ROLE_UPDATE
    #             - GUILD_ROLE_DELETE
    #             - CHANNEL_CREATE
    #             - CHANNEL_UPDATE
    #             - CHANNEL_DELETE
    #             - CHANNEL_PINS_UPDATE
    #             - THREAD_CREATE
    #             - THREAD_UPDATE
    #             - THREAD_DELETE
    #             - THREAD_LIST_SYNC
    #             - THREAD_MEMBER_UPDATE
    #             - THREAD_MEMBERS_UPDATE
    #             - STAGE_INSTANCE_CREATE
    #             - STAGE_INSTANCE_UPDATE
    #             - STAGE_INSTANCE_DELETE
    #         '''
    #         return (1 << 0)
    #     elif intents == 'GUILD_MEMBERS':
    #         '''
    #             - GUILD_MEMBER_ADD
    #             - GUILD_MEMBER_UPDATE
    #             - GUILD_MEMBER_REMOVE
    #             - THREAD_MEMBERS_UPDATE *
    #         '''
    #         return (1 << 1)
    #     elif intents == 'GUILD_BANS':
    #         '''
    #             - GUILD_BAN_ADD
    #             - GUILD_BAN_REMOVE
    #         '''
    #         return (1 << 2)
    #     elif intents == 'GUILD_EMOJIS_AND_STICKERS':
    #         '''
    #             - GUILD_EMOJIS_UPDATE
    #             - GUILD_STICKERS_UPDATE
    #         '''
    #         return (1 << 3)
    #     elif intents == 'GUILD_INTEGRATIONS': 
    #         '''
    #             - GUILD_INTEGRATIONS_UPDATE
    #             - INTEGRATION_CREATE
    #             - INTEGRATION_UPDATE
    #             - INTEGRATION_DELETE
    #         '''
    #         return (1 << 4)
    #     elif intents == 'GUILD_WEBHOOKS':
    #         '''
    #             - WEBHOOKS_UPDATE
    #         '''
    #         return (1 << 5)
    #     elif intents == 'GUILD_INVITES':
    #         '''
    #             - INVITE_CREATE
    #             - INVITE_DELETE
    #         '''
    #         return (1 << 6)
    #     elif intents == 'GUILD_VOICE_STATES': 
    #         '''
    #             - VOICE_STATE_UPDATE
    #         '''
    #         return (1 << 7)
    #     elif intents == 'GUILD_PRESENCES': 
    #         '''
    #             - PRESENCE_UPDATE
    #         '''
    #         return (1 << 8)
    #     elif intents == 'GUILD_MESSAGES':
    #         '''
    #             - MESSAGE_CREATE
    #             - MESSAGE_UPDATE
    #             - MESSAGE_DELETE
    #             - MESSAGE_DELETE_BULK
    #         '''
    #         return  (1 << 9)
    #     elif intents == 'GUILD_MESSAGE_REACTIONS':
    #         '''
    #             - MESSAGE_REACTION_ADD
    #             - MESSAGE_REACTION_REMOVE
    #             - MESSAGE_REACTION_REMOVE_ALL
    #             - MESSAGE_REACTION_REMOVE_EMOJI
    #         '''
    #         return  (1 << 10)
    #     elif intents == 'GUILD_MESSAGE_TYPING': 
    #         '''
    #             - TYPING_START
    #         '''
    #         return (1 << 11)
    #     elif intents == 'DIRECT_MESSAGES':
    #         '''
    #             - MESSAGE_CREATE
    #             - MESSAGE_UPDATE
    #             - MESSAGE_DELETE
    #             - CHANNEL_PINS_UPDATE
    #         '''
    #         return (1 << 12)
    #     elif intents == 'DIRECT_MESSAGE_REACTIONS': 
    #         '''
    #             - MESSAGE_REACTION_ADD
    #             - MESSAGE_REACTION_REMOVE
    #             - MESSAGE_REACTION_REMOVE_ALL
    #             - MESSAGE_REACTION_REMOVE_EMOJI
    #         '''
    #         return (1 << 13)
    #     elif intents == 'DIRECT_MESSAGE_TYPING': 
    #         '''
    #             - TYPING_START
    #         '''
    #         return (1 << 14)
    #     elif intents == 'GUILD_SCHEDULED_EVENTS': 
    #         '''
    #             - GUILD_SCHEDULED_EVENT_CREATE
    #             - GUILD_SCHEDULED_EVENT_UPDATE
    #             - GUILD_SCHEDULED_EVENT_DELETE
    #             - GUILD_SCHEDULED_EVENT_USER_ADD
    #             - GUILD_SCHEDULED_EVENT_USER_REMOVE
    #         '''
    #         return (1 << 16)
    #     else:
    #         raise Exception('Uknown Intent!')





class User(object):
    '''Contains a collection of user related asynchron methods.'''

    async def get_current_user(self) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-current-user'''
        return await self._request(method='GET', uri=f'/users/@me')

    async def get_user(self, user_id) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-user'''
        return await  self._request(method='GET', uri=f'/users/{user_id}')

    async def modify_current_user(self, username: str=None, avatar=None) -> dict:
        '''https://discord.com/developers/docs/resources/user#modify-current-user'''
        payload: dict = { 'username': username, 'avatar': avatar}
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request(method='PATCH', params=payload, uri=f'/users/@me')

    async def get_current_user_guilds(self, before=None, after=None, limit: int=None) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-current-user-guilds'''
        payload: dict = { 
            'before': before,
            'after': after,
            'limit': limit,
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request(method='GET', params=payload, uri=f'/users/@me/guilds')

    async def get_current_user_guild_member(self, guild_id) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-current-user-guild-member'''
        return await  self._request(method='GET', uri=f'/users/@me/guilds/{guild_id}/member')

    async def leave_guild(self, guild_id) -> dict:
        '''https://discord.com/developers/docs/resources/user#leave-guild'''
        return await  self._request(method='DELETE', uri=f'/users/@me/guilds/{guild_id}')

    async def create_dm(self, recipient_id) -> dict:
        '''https://discord.com/developers/docs/resources/user#create-dm'''
        payload: dict = { 'recipient_id': recipient_id }
        return await  self._request(method='POST', params=payload, uri=f'/users/@me/channels')

    async def create_group_dm(self, access_tokens: list, nicks: dict) -> dict:
        '''https://discord.com/developers/docs/resources/user#create-group-dm'''
        payload: dict = {
            'access_tokens': access_tokens,
            'nicks': nicks
        }
        return await  self._request(method='POST', params=payload, uri=f'/users/@me/channels')

    async def get_user_connections(self) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-user-connections'''
        return await  self._request(method='GET', uri=f'/users/@me/connections')

    async def list_voice_regions(self) -> dict:
        '''https://discord.com/developers/docs/resources/voice#list-voice-regions'''
        return await  self._request(method='GET', uri=f'/voice/regions')

    async def get_invite(self, invite_code, with_counts: bool=None, with_expiration: bool=None, guild_scheduled_event_id=None) -> dict:
        '''https://discord.com/developers/docs/resources/invite#get-invite'''
        payload: dict = {
            'with_counts': with_counts,
            'with_expiration': with_expiration,
            'guild_scheduled_event_id': guild_scheduled_event_id
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request(method='GET', params=payload, uri=f'/invites/{invite_code}')

    async def delete_invite(self, invite_code) -> dict:
        '''https://discord.com/developers/docs/resources/invite#delete-invite'''
        return await  self._request(method='DELETE', params=payload, uri=f'/invites/{invite_code}')



    




class Attachment(object):

    def __init__(self, 
        id: int, 
        filename: str=None, 
        description: str=None,
        content_type: str=None,
        size: int=None,
        url: str=None,
        proxy_url: str=None,
        height: int=None,
        width: int=None,
        ephemeral: bool=None
    ):
        self.id = id
        self.filename = filename
        self.description = description
        self.content_type = content_type
        self.size = size
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width
        self.ephemeral = ephemeral

    def asdict(self) -> dict:
        return {
            'id': self.id, 
            'filename': self.filename,
            'description': self.description,
            'content_type': self.content_type,
            'size': self.size,
            'url': self.url,
            'proxy_url': self.proxy_url,
            'height': self.height,
            'wisth': self.width,
            'ephemeral': self.ephemeral
        }



class Channel(object):
    '''Contains a collection of channel related methods.'''

    async def get_channel(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-channel'''
        return await self._request('GET', uri=f'/channels/{channel_id}')

    async def modify_channel(self, 
        channel_id,
        name: str=None,     
        icon=None, 
        type: int=None,
        position: int=None,
        topic: str=None,
        nsfw: bool=None,
        rate_limit_per_user: int=None,
        bitrate: int=None,
        user_limit: int=None,
        permission_overwrites: list=None,
        parent_id=None,
        rtc_region: str=None,
        video_quality_mode: int=None,
        default_auto_archive_duration: int=None,
        archived: bool=None,
        auto_archive_duration: int=None,
        locked: bool=None,
        invitable: bool=None,
        **kwargs
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#modify-channel
            
            ----- P A R M E T E R S -----
            If Channel == Group DM
                @name
                @icon
            If Channel == Guild Channel
                @name
                @type
                @position
                @topic
                @nsfw
                @rate_limit_per_user
                @bitrate
                @user_limit
                @permission_overwrites
                @parent_id
                @rtc_region
                @video_quality_mode
                @default_auto_archive_duration
            If Chanel == Thread
                @name
                @archived
                @auto_archive_duration
                @locked
                @invitable
                @rate_limit_per_user
        '''
        payload: dict = {
            'name': name,
            'icon': icon,
            'type': type,
            'topic': topic,
            'nsfw': nsfw,
            'rate_limit_per_user': rate_limit_per_user,
            'bitrate': bitrate,
            'user_limit': user_limit,
            'permission_overwrites': permission_overwrites,
            'parent_id': parent_id,
            'rtc_region': rtc_region,
            'video_quality_mode': video_quality_mode,
            'default_auto_archive_duration': default_auto_archive_duration,
            'archived': archived,
            'auto_archive_duration': auto_archive_duration,
            'locked': locked,
            'invitable': invitable
        }        
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('PATCH', params=payload, uri=f'/channels/{channel_id}')
        
    async def delete_close_channel(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#deleteclose-channel'''
        return await  self._request('DELETE', uri=f'/channels/{channel_id}')

    async def get_channel_messages(self,
        channel_id,
        around=None,
        before=None,
        after=None,
        limit: int=None,
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-channel-messages'''
        payload: dict = {
            'around': around,
            'before': before,
            'after': after,
            'limit': limit
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('GET', params=payload, uri=f'/channels/{channel_id}/messages')

    async def get_channel_message(self) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-channel-message'''
        return await  self._request('GET',uri=f'/channels/{channel_id}/messages')

    async def create_message(self,
        channel_id,
        content: str=None,
        tts: bool=None,
        embeds: [dict]=None,
        allowed_mentions=None,
        message_reference=None,
        components: list=None,
        files: [str]=None,
        attachments: [dict]=None,
        sticker_ids: list=None,
        flags: int=None,
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#create-message'''

        payload: dict = {
            'content': content,
            'tts': tts,
            'allowed_mentions': allowed_mentions,
            'message_reference': message_reference,
            'components': components,
            'sticker_ids': sticker_ids,
            'flags': flags,
            'embeds': embeds,
            'attachments': attachments
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        uri = f'/channels/{channel_id}/messages'
        if files != None:
            return await  self._send_file_attachment(method='POST', uri=uri, file_names=files, payload=payload)
        else:
            return await  self._request(method='POST', uri=uri, params=payload)

    async def crosspost_message(self, message_id, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#crosspost-message'''
        return await self._request('POST', uri=f'/channels/{channel_id}/messages/{message_id}/crosspost')

    async def create_reaction(self, channel_id, message_id, emoji) -> dict:
        '''https://discord.com/developers/docs/resources/channel#create-reaction
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        return await self._request('PUT', uri=f'/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me')

    async def delete_own_reaction(self, channel_id, message_id, emoji) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-own-reaction
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        return await self._request('DELETE', uri=f'/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me')

    async def delete_user_reaction(self, channel_id, message_id, user_id, emoji) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-user-reaction
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        return await self._request('DELETE', uri=f'/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/{user_id}')

    async def get_reactions(self, channel_id, message_id, emoji, after=None, limit: int=None) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-reactions
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        return await self._request('GET', uri=f'/channels/{channel_id}/messages/{message_id}/reactions/{emoji}')

    async def delete_all_reactions(self, channel_id, message_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-all-reactions'''
        return await self._request('DELETE', uri=f'/channels/{channel_id}/messages/{message_id}/reactions')
        
    async def delete_all_reactions_for_emoji(self, channel_id, message_id, emoji) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-all-reactions-for-emoji
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        return await self._request('DELETE', uri=f'/channels/{channel_id}/messages/{message_id}/reactions/{emoji}')

    async def edit_message(self,
        channel_id,
        message_id, 
        content: str=None,
        embeds: [dict]=None,
        flags: int=None,
        allowed_mentions=None,
        components=list,
        files: [str]=None,
        payload_json: str=None,
        attachments: list=None,
        **kwargs
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#edit-message'''
        payload: dict = {
            'content': content,
            'flags': flags,
            'allowed_mentions': allowed_mentions,
            'components': components,
            'payload_json': payload_json,
            'attachments': attachments,
            'embeds': embeds
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        uri = f'/channels/{channel_id}/messages/{message_id}'
        if files != None:
            return await  self._send_file_attachment(method='POST', uri=uri, file_names=files, payload=payload)
        else:
            return await  self._request(method='POST', uri=uri, payload=payload)

    async def delete_message(self, channel_id, message_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-message'''
        return await  self._request('DELETE', uri=f'/channels/{channel_id}/messages/{message_id}')

    async def bulk_delete_message(self, channel_id, messages: list) -> dict:
        '''https://discord.com/developers/docs/resources/channel#bulk-delete-messages'''
        paylaod = { 'messages': messages }
        return await  self._request('POST', params=payload, uri=f'/channels/{channel_id}/messages/bulk-delete')

    async def edit_channel_permissions(self, channel_id, overwrite_id, allow=None, deny=None, type: int=None) -> dict:
        '''https://discord.com/developers/docs/resources/channel#edit-channel-permissions'''
        payload: dict = {
            'allow': allow,
            'deny': deny,
            'type': type
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('PUT', params=payload, uri=f'/channels/{channel_id}/permissions/{overwrite_id}')

    async def get_channel_invites(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-channel-invites'''
        return await  self._request('GET', uri=f'/channels/{channel_id}/invites')

    async def create_channel_invite(self,
        channel_id,
        max_age: int=None,
        max_uses: int=None,
        temporary: bool=None,
        unique: bool=None,
        target_type: int=None,
        target_user_id=None,
        target_application_id=None,
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#create-channel-invite'''
        payload: dict = {
            'max_age': max_age,
            'max_uses': max_uses,
            'temporary': temporary,
            'unique': unique,
            'target_type': target_type,
            'target_user_id': target_user_id,
            'target_application_id': target_application_id
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('POST', params=payload, uri=f'/channels/{channel_id}/invites')

    async def delete_channel_permission(self, channel_id, overwrite_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-channel-permission'''
        return await  self._request('DELETE', uri=f'/channels/{channel_id}/permissions/{overwrite_id}')

    async def follow_news_channel(self, channel_id, webhook_channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#follow-news-channel'''
        payload: dict = { 'webhook_channel_id': webhook_channel_id }
        return await  self._request('POST', params=payload, uri=f'/channels/{channel_id}/followers')

    async def trigger_typing_indicatoe(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#trigger-typing-indicator'''
        return await  self._request('POST', uri=f'/channels/{channel_id}/typing')

    async def get_pinned_message(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-pinned-messages'''
        return await  self._request('GET', uri=f'/channels/{channel_id}/pins')

    async def pin_message(self, channel_id, message_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#pin-message'''
        return await  self._request('PUT', uri=f'/channels/{channel_id}/pins/{message_id}')

    async def unpin_message(self, channel_id,  message_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#unpin-message'''
        return await  self._request('DELETE', uri=f'/channels/{channel_id}/pins/{message_id}')

    async def group_dm_add_recipient(self, channel_id, user_id, access_token, nick) -> dict:
        '''https://discord.com/developers/docs/resources/channel#group-dm-add-recipient'''
        payload: dict = {
            'access_token': access_token,
            'nick':nick
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('PUT', params=payload, uri=f'/channels/{channel_id}/recipients/{user_id}')

    async def group_dm_remove_recipient(self, channel_id, user_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#group-dm-remove-recipient'''
        return await  self._request('DELETE', uri=f'/channels/{channel_id}/recipients/{user_id}')

    async def start_thread_with_message(self, channel_id, user_id, name: str, auto_archive_duration: int=None, rate_limit_per_user: int=None) -> dict:
        '''https://discord.com/developers/docs/resources/channel#start-thread-with-message'''
        payload: dict = { 
            'name': name,
            'auto_archive_duration': auto_archive_duration,
            'rate_limit_per_user': rate_limit_per_user
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('POST', params=payload, uri=f'/channels/{channel_id}/messages/{message_id}/threads')

    async def start_thread_without_message(self, 
        channel_id,
        auto_archive_duration: str=None,
        type: int=None,
        invitable: bool=None,
        rate_limit_per_user: int=None,
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#start-thread-without-message'''
        payload: dict = { 
            'name': name,
            'auto_archive_duration': auto_archive_duration,
            'type': type,
            'invitable': invitable,
            'rate_limit_per_user': rate_limit_per_user
        }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('POST', params=payload, uri=f'/channels/{channel_id}/threads')

    async def join_thread(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#join-thread'''
        return await  self._request('PUT', uri=f'/channels/{channel_id}/thread-members/@me')

    async def add_thread_member(self, channel_id, user_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#add-thread-member'''
        return await  self._request('PUT', uri=f'/channels/{channel_id}/thread-members/{user_id}')

    async def leave_thread(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#leave-thread'''
        return await  self._request('DELETE', uri=f'/channels/{channel_id}/thread-members/@me')

    async def remove_thread_member(self, channel_id, user_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#remove-thread-member'''
        return await  self._request('DELETE', uri=f'/channels/{channel_id}/thread-members/{user_id}')

    async def get_thread_member(self, channel_id, user_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-thread-member'''
        return await  self._request('GET', uri=f'/channels/{channel_id}/thread-members/{user_id}')

    async def list_thread_member(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-thread-members'''
        return await  self._request('GET', uri=f'/channels/{channel_id}/thread-members')

    async def list_active_threads(self, channel_id, threads: list, members: list, has_more: bool ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-active-threads'''
        payload: dict = {
            'threads': threads,
            'members': members,
            'has_more': has_more
        }
        return await  self._request('GET', params=payload, uri=f'/channels/{channel_id}/threads/active')

    async def list_public_archived_threads(self, channel_id, before=None, limit:int=None ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-public-archived-threads'''
        payload: dict = { 'before': before, 'limit': limit}
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('GET', params=params, uri=f'/channels/{channel_id}/threads/archived/public')

    async def list_private_archived_threads(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-private-archived-threads'''
        payload: dict = { 'before': before, 'limit': limit }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('GET', params=payload, uri=f'/channels/{channel_id}/threads/archived/private')

    async def list_joined_private_archived_threads(self, channel_id, before=None, limit:int=None) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-joined-private-archived-threads'''
        payload: dict = { 'before': before, 'limit': limit }
        payload: dict = {k:v for k,v in payload.items() if v is not None}
        return await  self._request('GET', params=payload, uri=f'/channels/{channel_id}/users/@me/threads/archived/private')


import numpy as np

class Permissions(object):
    '''Object to store and access User and Bot Permissions'''

    _roles = [
        'CREATE_INSTANT_INVITE', 'KICK_MEMBERS', 'BAN_MEMBERS', 'ADMINISTRATOR'
        'MANAGE_CHANNELS', 'MANAGE_GUILD','ADD_REACTIONS', 'VIEW_AUDIT_LOG',
        'PRIORITY_SPEAKER', 'STREAM', 'VIEW_CHANNEL',
        'SEND_MESSAGES', 'SEND_TTS_MESSAGES', 'MANAGE_MESSAGES',
        'EMBED_LINKS', 'ATTACH_FILES', 'READ_MESSAGE_HISTORY',
        'MENTION_EVERYONE', 'USE_EXTERNAL_EMOJIS', 'VIEW_GUILD_INSIGHTS',
        'CONNECT', 'SPEAK', 'MUTE_MEMBERS', 'DEAFEN_MEMBERS', 
        'MOVE_MEMBERS', 'USE_VAD','CHANGE_NICKNAME', 'MANAGE_NICKNAMES',
        'MANAGE_ROLES', 'MANAGE_WEBHOOKS', 'MANAGE_EMOJIS_AND_STICKERS',
        'USE_APPLICATION_COMMANDS', 'REQUEST_TO_SPEAK', 'MANAGE_EVENTS', 'MANAGE_THREADS',
        'CREATE_PUBLIC_THREADS', 'CREATE_PRIVATE_THREADS', 'USE_EXTERNAL_STICKERS',
        'SEND_MESSAGES_IN_THREADS', 'START_EMBEDDED_ACTIVITIES', 'MODERATE_MEMBERS',
    ]

    def list_roles(self) -> [str]:
        return self._roles

    def get_role(self, role_name: str='') -> int:
        for i, role in enumerate(self._roles):
            if role == role_name:
                return (1<<i)
        raise ValueError(f'Role {role_name} not found!')
    
    def get_roles(self, roles_list: [str]) -> int:
        return np.sum([self.get_role(role) for role in roles_list])




# import asyncio
# import websockets as ws
# import sys, traceback
# import json
# import time 
# import logging 
# import numpy as np

# THIS IS OLD AND SHOULD NOT BE USED ANYMORE!
# import requests
# class WSClient(object):
#     '''
#         Connect to the Discord websocket-feed.

#         @param: token (str): required 
#         @param: url (str):  required
#     '''
#     _heartbeat_interval: int = int(41250 / 1000) # in seconds
#     _last_ping = None
#     _sequence = None

#     all_intents: dict = [
#         'GUILDS', 'GUILD_MEMBERS', 'GUILD_BANS', 'GUILD_EMOJIS_AND_STICKERS', 'GUILD_INTEGRATIONS', 'GUILD_WEBHOOKS',
#         'GUILD_INVITES', 'GUILD_VOICE_STATES', 'GUILD_PRESENCES', 'GUILD_MESSAGES', 'GUILD_MESSAGE_REACTIONS',
#         'GUILD_MESSAGE_TYPING', 'DIRECT_MESSAGES','DIRECT_MESSAGE_REACTIONS', 'GUILD_SCHEDULED_EVENTS', 'GUILD_SCHEDULED_EVENTS'
#     ]

#     _url: str = 'wss://gateway.discord.gg/?v=9&encording=json'

#     def __init__(self, token: str, intents: list, callback=None, url: str=None):
#         self._log = logging.getLogger(__name__)

#         self._token: str = token
#         self._intents: [str] = np.sum([self._get_intents(intent) for intent in intents])

#         if url != None:
#             self._url: str = url

#         self._callback = callback
#         self._connect()

#     async def _run(self) -> None:
#         try:
#             keepwaiting: bool = True
#             async with ws.connect(self._url) as socket:
#                 self.ws = socket
#                 await self._authenticate() 

#                 try:
#                     while keepwaiting:
#                         if not self._last_ping or int(time.time()) - self._last_ping > self._get_ws_pingtimeout():
#                             await self._send_ping()
#                         try:
#                             evt = await asyncio.wait_for(self.ws.recv(), timeout=self._get_ws_pingtimeout())
#                         except asyncio.TimeoutError:
#                             self._log.info(f'No message in {self._get_ws_pingtimeout()} seconds')
#                             await self._send_ping()
#                         except asyncio.CancelledError:
#                             self._log.warn(f'Cancelled error')
#                             await self.ws.ping()
#                         else:
#                             try:
#                                 evt_obj = json.loads(evt)
#                                 if evt_obj['op'] != 11 and 's' in evt_obj and evt_obj['s']: # setting the _sequence
#                                     self._sequence = evt_obj['s']
#                                     if evt_obj['op'] == 10:  
#                                         self._heartbeat_interval = int(evt_obj['d']['heartbeat_interval']) / 1000
#                             except ValueError:
#                                 pass
#                             else:
#                                 if self._callback:
#                                     await self._callback(evt_obj)
                    
#                 except ws.ConnectionClosed:
#                     keep_waiting = False
#                     self._log.warn(f'Websocket connection closed!  {e} {traceback.format_exc()}')
#                     await self._reconnect()
#                 except Exception as e:
#                     self._log.error(f'WS exception: {e} {traceback.format_exc()}')
#                     keep_waiting = False
#                     await self._reconnect()

#         except Exception as e:
#             self._log.error(f'Exception!: {e} {traceback.format_exc()}')


#     def _connect(self) -> None:
#         self._log.info('Connecting...')
#         self._conn = asyncio.ensure_future(self._run())

#     async def _reconnect(self) -> None:
#         self._log.info('Reconnecting Websocket...')
#         await self._close_connection()
#         self._connect()
    
#     async def _close_connection(self) -> None:
#         await self.ws.close()

#     async def send_message(self, message: dict) -> None:
#         await self.ws.send(json.dumps(message))

#     async def _send_ping(self):
#         self._log.info(f'Sending ping now, _sequence: {self._sequence}')
#         await self.send_message({
#             'op': 1,
#             'd': self._sequence
#         })
#         self._last_ping = int(time.time())
    
#     async def _authenticate(self) -> None:
#         await self.send_message(self._get_payload(kind='init'))

#     def _get_ws_pingtimeout(self) -> None:
#         return self._heartbeat_interval

#     def _get_payload(self, kind: str) -> dict:
#         '''https://discord.com/developers/docs/topics/gateway#identifying'''
#         op: int = None
#         if kind == 'reconnect':
#             op = 7
#         elif kind == 'resume':
#             op = 6
#         elif kind == 'init':
#             op = 2
#         else:
#             raise ValueError
#         return {
#             'op': op,
#             'd': {
#                 'token': self._token,
#                 'intents': f'{self._intents}',
#                 'properties': {
#                     '$os': sys.platform,
#                     '$browser': 'disco',
#                     '$device': 'pc'
#                 },
#                 'compress': True
#             }
#         }
#     @classmethod
#     def get_intents_list(cls) -> list:
#         return cls.all_intents

#     def _get_intents(self, intents: str) -> int:
#         # https://discord.com/developers/docs/topics/gateway#gateway-intents
#         if intents == 'GUILDS':
#             '''
#                 - GUILD_CREATE
#                 - GUILD_UPDATE
#                 - GUILD_DELETE
#                 - GUILD_ROLE_CREATE
#                 - GUILD_ROLE_UPDATE
#                 - GUILD_ROLE_DELETE
#                 - CHANNEL_CREATE
#                 - CHANNEL_UPDATE
#                 - CHANNEL_DELETE
#                 - CHANNEL_PINS_UPDATE
#                 - THREAD_CREATE
#                 - THREAD_UPDATE
#                 - THREAD_DELETE
#                 - THREAD_LIST_SYNC
#                 - THREAD_MEMBER_UPDATE
#                 - THREAD_MEMBERS_UPDATE
#                 - STAGE_INSTANCE_CREATE
#                 - STAGE_INSTANCE_UPDATE
#                 - STAGE_INSTANCE_DELETE
#             '''
#             return (1 << 0)
#         elif intents == 'GUILD_MEMBERS':
#             '''
#                 - GUILD_MEMBER_ADD
#                 - GUILD_MEMBER_UPDATE
#                 - GUILD_MEMBER_REMOVE
#                 - THREAD_MEMBERS_UPDATE *
#             '''
#             return (1 << 1)
#         elif intents == 'GUILD_BANS':
#             '''
#                 - GUILD_BAN_ADD
#                 - GUILD_BAN_REMOVE
#             '''
#             return (1 << 2)
#         elif intents == 'GUILD_EMOJIS_AND_STICKERS':
#             '''
#                 - GUILD_EMOJIS_UPDATE
#                 - GUILD_STICKERS_UPDATE
#             '''
#             return (1 << 3)
#         elif intents == 'GUILD_INTEGRATIONS': 
#             '''
#                 - GUILD_INTEGRATIONS_UPDATE
#                 - INTEGRATION_CREATE
#                 - INTEGRATION_UPDATE
#                 - INTEGRATION_DELETE
#             '''
#             return (1 << 4)
#         elif intents == 'GUILD_WEBHOOKS':
#             '''
#                 - WEBHOOKS_UPDATE
#             '''
#             return (1 << 5)
#         elif intents == 'GUILD_INVITES':
#             '''
#                 - INVITE_CREATE
#                 - INVITE_DELETE
#             '''
#             return (1 << 6)
#         elif intents == 'GUILD_VOICE_STATES': 
#             '''
#                 - VOICE_STATE_UPDATE
#             '''
#             return (1 << 7)
#         elif intents == 'GUILD_PRESENCES': 
#             '''
#                 - PRESENCE_UPDATE
#             '''
#             return (1 << 8)
#         elif intents == 'GUILD_MESSAGES':
#             '''
#                 - MESSAGE_CREATE
#                 - MESSAGE_UPDATE
#                 - MESSAGE_DELETE
#                 - MESSAGE_DELETE_BULK
#             '''
#             return  (1 << 9)
#         elif intents == 'GUILD_MESSAGE_REACTIONS':
#             '''
#                 - MESSAGE_REACTION_ADD
#                 - MESSAGE_REACTION_REMOVE
#                 - MESSAGE_REACTION_REMOVE_ALL
#                 - MESSAGE_REACTION_REMOVE_EMOJI
#             '''
#             return  (1 << 10)
#         elif intents == 'GUILD_MESSAGE_TYPING': 
#             '''
#                 - TYPING_START
#             '''
#             return (1 << 11)
#         elif intents == 'DIRECT_MESSAGES':
#             '''
#                 - MESSAGE_CREATE
#                 - MESSAGE_UPDATE
#                 - MESSAGE_DELETE
#                 - CHANNEL_PINS_UPDATE
#             '''
#             return (1 << 12)
#         elif intents == 'DIRECT_MESSAGE_REACTIONS': 
#             '''
#                 - MESSAGE_REACTION_ADD
#                 - MESSAGE_REACTION_REMOVE
#                 - MESSAGE_REACTION_REMOVE_ALL
#                 - MESSAGE_REACTION_REMOVE_EMOJI
#             '''
#             return (1 << 13)
#         elif intents == 'DIRECT_MESSAGE_TYPING': 
#             '''
#                 - TYPING_START
#             '''
#             return (1 << 14)
#         elif intents == 'GUILD_SCHEDULED_EVENTS': 
#             '''
#                 - GUILD_SCHEDULED_EVENT_CREATE
#                 - GUILD_SCHEDULED_EVENT_UPDATE
#                 - GUILD_SCHEDULED_EVENT_DELETE
#                 - GUILD_SCHEDULED_EVENT_USER_ADD
#                 - GUILD_SCHEDULED_EVENT_USER_REMOVE
#             '''
#             return (1 << 16)
#         else:
#             raise Exception('Uknown Intent!')




class Embed(object):

    def __init__(self,
        title: str=None,
        type: str=None,
        description: str=None,
        url: str=None,
        timestamp=None,
        color: int=None,
        footer=None,
        image=None,
        thumbnail=None,
        video=None,
        provider=None,
        author: str=None,
        fields: [dict]=None,
    ):
    
        self.title = title
        self.type = type
        self.description = description
        self.url = url
        self.timestamp = timestamp
        self.color = color
        self.footer = footer
        self.image = image
        self.thumbnail = thumbnail
        self.video = video
        self.provider = provider
        self.author = author
        self.fields = fields

    def asdict(self) -> dict:
        return {
            'title': self.title,
            'type': self.type,
            'description': self.description,
            'url': self.url,
            'timestamp': self.timestamp,
            'color': self.color,
            'footer': self.footer,
            'image': self.image,
            'thumbnail': self.thumbnail,
            'video': self.video,
            'provider': self.provider,
            'author': self.author,
            'fields': self.fields
        }
